/*
Mike Morris and Karl Frinkle, Southeastern Oklahoma State University
Last Modified: 2012.09.25

This program will compute the product of two square matrices
of the same size, and store the result in a new binary file.
Each matrix should be stored in a binary file, and the entries are 
to be integer entries. The two programs MatrixMultMPIGenerator and
MatrixMultMPIGeneratorDiagonal, located in the 00_MatrixCreation
directory, can be used to make binary matrix files, the latter will make 
diagonal matrices, for verification of code purposes.  These two are
MPI programs, as opposed to the C programs to make binary matrix files.

All of the read/writes done in this program are MPI, notice how this
program runs.  The matrices are hard coded into this program, so you 
will have to change the file names, and the value of n below if you 
wish to use different sized matrices.  This program uses an 'even
distribution' technique, where each node just gets a chunk of the
entries to work on, and all nodes read/write with MPI to the same file
at the same time.
*/

#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include <stdlib.h>
#include <memory.h>

int main(int argc, char** argv)
{
	char fnameA[] = "A1000.bin"; //change these if you wish to try
	char fnameB[] = "B1000.bin"; //a different dimension matrix
	char fnameC[] = "C1000.bin"; //multiplication

	char nodename[80]; //for name call of nodes
	int lengthname;

	int n = 1000*1000; //The number of entries in the matrices to
		   //multiply together, relate this to fnameA,B,C above

	int N = sqrt(n); //gives the number of rows/columns
	
	int c, r, s, t, k;

	int rowIdx = 0; //index of the row, default to zero before loop
	int colIdx = 0; //index of the row, default to zero before loop

	int *bufc;
	int *bufb;
	int *bufa;

	int acc; //accumulation of multiplcation of row r of A by column c of B
	int my_rank, world_size, destination, tag, source, length;
	
	int L, R;

	double starttime;
	double endtime;

	MPI_Status status;
	MPI_File fpA;
	MPI_File fpB;
	MPI_File fpC;
	MPI_Request req;
	MPI_Offset offset;

	MPI_Init(&argc, &argv); /* note that argc and argv are passed by address */
	
	MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
	MPI_Comm_size(MPI_COMM_WORLD,&world_size);

	int chunk_size = n/world_size;

	bufa = (int *)malloc(n*sizeof(int));
	bufb = (int *)malloc(n*sizeof(int));
	bufc = (int *)malloc(n*sizeof(int));
	
	offset = n*sizeof(int);

	MPI_File_open(MPI_COMM_WORLD, fnameA, MPI_MODE_RDONLY, MPI_INFO_NULL, &fpA);	
	MPI_File_open(MPI_COMM_WORLD, fnameB, MPI_MODE_RDONLY, MPI_INFO_NULL, &fpB);
	MPI_File_open(MPI_COMM_WORLD, fnameC, MPI_MODE_RDWR, MPI_INFO_NULL, &fpC);

	MPI_File_set_size(fpC, offset);
	MPI_Get_processor_name(nodename, &lengthname);

	L = chunk_size*my_rank;
	if (my_rank != world_size-1)
	{
		R = (my_rank+1)*chunk_size -1;
	}
	else
	{
		R = n-1;
	}

	printf("Node %d initializing, will be working on entries %d through %d. \n", my_rank, L, R);


	for(k=L; k<=R; k++){

		acc=0;
		c = fmod(k,N)+1;
		r = floor(k/N)+1;
		bufa[0] = 0;
		bufb[0] = 0;
		bufc[0] = 0;
		
		for(s=0; s<N; s++){
			rowIdx = (s+(r-1)*N)*sizeof(int);
			colIdx = (s*N+(c-1))*sizeof(int);
			MPI_File_read_at(fpA, rowIdx, bufa, 1, MPI_INT, &status);
			MPI_File_read_at(fpB, colIdx, bufb, 1, MPI_INT, &status);
			acc = acc + bufa[0]*bufb[0];
		}
		bufc[0] = acc;
		MPI_File_write_at(fpC, k*sizeof(int), bufc, 1, MPI_INT, &status);
		if(c == r) {
			printf("[%s]: Node %d, for entry %d computed a value of %d\n", nodename, my_rank, k,  acc);
		}
	}
	MPI_File_close(&fpA);
	MPI_File_close(&fpB);
	MPI_File_close(&fpC);
	MPI_Finalize();
}
