/*
Mike Morris and Karl Frinkle, Southeastern Oklahoma State University
Last Modified: 2012.09.25

This program will compute the product of two square matrices
of the same size, and store the result in a new binary file.
Each matrix should be stored in a binary file, and the entries are 
to be integer entries. The two programs MatrixMultMPIGenerator and
MatrixMultMPIGeneratorDiagonal, located in the 00_MatrixCreation
directory, can be used to make binary matrix files, the latter will make 
diagonal matrices, for verification of code purposes.  These two are
MPI programs, as opposed to the C programs to make binary matrix files.

All of the read/writes done in this program are done not with MPI, but
with standard C, and only on the head node.  Compare runtime with this
code and the MatrixMultRA program from the even distribution directory
03_BINARY_MultiplyEvenDistribution.

This program has one runtime variable, which is the row/column dimension
of the matrices to be multiplied together.  For instance, if you have
A100.bin and B100.bin in your working directory, then these are two 100 by
100 matrices.  To run this, you would execute:

time mpirun -np 12 -machinefile ~/machines MatrixMultNodeDist 100

You must use more than two processors for this program to run correctly.
*/

#include <stdio.h>
#include <math.h>
#include <mpi.h>
#include <stdlib.h>
#include <memory.h>

#define REQUEST 1
#define REPLY 2
#define FINAL 3

int main(int argc, char ** argv){

	int i = 0;

	int N = atoi(argv[1]); //gives the number of rows/columns
	int n = N * N; // number of entries in the matrix

	char nodename[80]; //for getting name of each node
	int lengthname;

	int j=0; //these will be used as counters in for loops
	int d, kk; //value that gets set to j for each node 

	int c, r, s, t, k;

	int *buf;
	int *bufans;
	int *element;

	int rowIdx = 0; //index of the row, default to zero before loop
	int colIdx = 0; //index of the row, default to zero before loop

	int counter=0;

	int acc; //accumulation of multiplcation of row r of A by column c of B

	int my_rank, world_size, destination, tag, source, length;

	int request = 1;

	char fnameA[80];
	strcpy(fnameA, "A");
	strcat(fnameA, argv[1]);
	strcat(fnameA, ".bin");

	char fnameB[80];
	strcpy(fnameB, "B");
	strcat(fnameB, argv[1]);
	strcat(fnameB, ".bin");

	char fnameC[80];
	strcpy(fnameC, "C");
	strcat(fnameC, argv[1]);
	strcat(fnameC, ".bin");

	FILE *fptB;
	FILE *fptBT;

	char fnameBT[80];
	strcpy(fnameBT, "BT");
	strcat(fnameBT, argv[1]);
	strcat(fnameBT, ".txt");	

	MPI_Status status;
	MPI_File fpA;
	MPI_File fpB;
	MPI_File fpC;
	MPI_Offset offset;

	int position;

	MPI_Init(&argc, &argv); /* note that argc and argv are passed by address */
	
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
	MPI_Comm_size(MPI_COMM_WORLD, &world_size);

	buf = (int *)malloc((2*N+1)*sizeof(int));

	bufans = (int *)malloc(2*sizeof(int));

	element = (int *)malloc(N*sizeof(int));
	
	offset = n*sizeof(int);

	MPI_File_open(MPI_COMM_WORLD, fnameA, MPI_MODE_RDONLY, MPI_INFO_NULL, &fpA);	
	MPI_File_open(MPI_COMM_WORLD, fnameC, MPI_MODE_RDWR, MPI_INFO_NULL, &fpC);

	MPI_File_set_size(fpC, offset);


	MPI_Get_processor_name(nodename, &lengthname);

	if(my_rank == 0){


//	 	Here we perform the transpose of B before we get started on anything else...
	
		printf("We are performing the tranpose of B\n");
	
		fptB = fopen(fnameB, "rb");
		fptBT = fopen(fnameBT, "wb");
	
		int cntr = 0;
		
		int jj;	
	
		for(jj=0; jj<=N-1;jj++){
			for(kk=0; kk<=N-1;kk++){
				fseek(fptB, (kk*N+jj)*4, 0);
				fread(&element[kk], 4, 1, fptB);	
			}	
			fseek(fptBT, cntr*4*N, SEEK_SET);
			fwrite(element, 4, N, fptBT);
			cntr++;
		}
		fclose(fptB);
		fclose(fptBT);

		printf("Transpose of B complete\n");
	}

	MPI_File_open(MPI_COMM_WORLD, fnameBT, MPI_MODE_RDONLY, MPI_INFO_NULL, &fpB);

	if(my_rank == 0){
		printf("Master node (%d) says hi \n", my_rank);
		while(request){
			MPI_Recv(&request, 1, MPI_INT, MPI_ANY_SOURCE, REQUEST, MPI_COMM_WORLD, &status);
//			printf("Master node (%d), received request for information from slave node (%d) \n", my_rank, status.MPI_SOURCE);		

			buf[0] = j;
			c = fmod(j,N)+1;
			r = floor(j/N)+1;
			
			rowIdx = ((r-1)*N)*sizeof(int);
			MPI_File_read_at(fpA, rowIdx, &buf[1], N, MPI_INT, &status);

			colIdx = ((c-1)*N)*sizeof(int);				
			position = N+1;
			MPI_File_read_at(fpB, colIdx, &buf[position], N, MPI_INT, &status);

			MPI_Send(buf, 2*N+1, MPI_INT, status.MPI_SOURCE, REPLY, MPI_COMM_WORLD);
//			printf("Master node (%d) - loop %d  - sending information to node (%d)\n", my_rank, j, status.MPI_SOURCE);
			if(j==n-1){
				request = 0;
				j++;
				buf[0] = j;
//				printf("Master node (%d) - changing request at loop %d to %d\n", my_rank, j, request);
				for(kk=2; kk<world_size; kk++){
					MPI_Send(buf, 2*N+1, MPI_INT, kk, REPLY, MPI_COMM_WORLD);
				}
			}
			j++;
			if(fmod(j,N)==0) printf("Loop number is %d, program is %3.2f percent done.\n",j, 100*(float)j/(float)n);
		}
	}
	else if (my_rank==1){
		while(counter<n){
			MPI_Recv(bufans, 2, MPI_INT, MPI_ANY_SOURCE, FINAL, MPI_COMM_WORLD, &status);
			counter++;
//			printf("Write node (%d), received, for the %d time, the value %d for entry %d from slave node (%d) \n", my_rank, counter, bufans[1], bufans[0], status.MPI_SOURCE);	
			MPI_File_write_at(fpC, bufans[0]*sizeof(int), &bufans[1], 1, MPI_INT, &status);
		}
	}
	else{
		tag = my_rank;
		while(j<n){
//			printf("Node (%d), sending request for information to master node (0) \n", my_rank);
			MPI_Send(&request, 1, MPI_INT, 0, REQUEST, MPI_COMM_WORLD);
			MPI_Recv(buf, 2*N+1, MPI_INT, 0, REPLY, MPI_COMM_WORLD, &status);
			counter ++;
			j = buf[0];
			if(j<n){
				acc = 0;
				for(kk=1;kk<N+1;kk++){
					acc = acc+buf[kk]*buf[kk+N];
				}
				bufans[0] = j;
				bufans[1] = acc;

//				printf("Node %d computed entry %d with a value of %d\n", my_rank, bufans[0], bufans[1]);

				MPI_Send(bufans, 2, MPI_INT, 1, FINAL, MPI_COMM_WORLD);
			}
		}	
		printf("\n[%s]: Node (%d), received %d entries to work on, from master node (0)\n", nodename, my_rank, counter-1);
	}

	MPI_File_close(&fpA);
	MPI_File_close(&fpB);
	MPI_File_close(&fpC);

	if(my_rank == 0){
		remove(fnameBT);
	}

	MPI_Finalize();
}
